/**
 * @fileoverview Forbid the use of equality comparisons in assert.ok/notOk.
 * @author Kevin Partington
 */
"use strict";

const utils = require("../utils");

//------------------------------------------------------------------------------
// Rule Definition
//------------------------------------------------------------------------------

module.exports = {
    meta: {
        docs: {
            description: "forbid equality comparisons in assert.ok/assert.notOk",
            category: "Best Practices"
        },
        messages: {
            noEqualityCheckInOk: "Unexpected equality comparison in {{assertion}} call. Use {{suggestion}}({{a}}, {{b}}) instead."
        },
        schema: [
            {
                type: "object",
                properties: {
                    allowGlobal: {
                        type: "boolean"
                    }
                },
                additionalProperties: false
            }
        ]
    },

    create: function (context) {
        // Declare a stack in case of nested test cases (not currently supported
        // in QUnit).
        const asyncStateStack = [],
            DEFAULT_OPTIONS = {
                allowGlobal: true
            },
            options = context.options[0] || DEFAULT_OPTIONS,
            sourceCode = context.getSourceCode();

        function getAssertContextVar() {
            const state = asyncStateStack[asyncStateStack.length - 1];
            return state && state.assertContextVar;
        }

        function isOk(calleeNode) {
            const assertContextVar = getAssertContextVar();

            const isOk = calleeNode.type === "Identifier" && calleeNode.name === "ok";

            const isAssertOk = calleeNode.type === "MemberExpression" &&
                calleeNode.object.type === "Identifier" &&
                calleeNode.object.name === assertContextVar &&
                calleeNode.property.type === "Identifier" &&
                calleeNode.property.name === "ok";

            if (options.allowGlobal) {
                return isOk || isAssertOk;
            }

            return isAssertOk;
        }

        function isNotOk(calleeNode) {
            const assertContextVar = getAssertContextVar();

            const isNotOk = calleeNode.type === "Identifier" && calleeNode.name === "notOk";

            const isAssertNotOk = calleeNode.type === "MemberExpression" &&
                calleeNode.object.type === "Identifier" &&
                calleeNode.object.name === assertContextVar &&
                calleeNode.property.type === "Identifier" &&
                calleeNode.property.name === "notOk";

            if (options.allowGlobal) {
                return isNotOk || isAssertNotOk;
            }

            return isAssertNotOk;
        }

        function isOkOrNotOk(calleeNode) {
            return isOk(calleeNode) || isNotOk(calleeNode);
        }

        function isEqual(arg) {
            return arg.type === "BinaryExpression" &&
                ["===", "==", "!==", "!="].indexOf(arg.operator) !== -1;
        }

        function isStrict(arg) {
            return arg.type === "BinaryExpression" &&
                ["===", "!=="].indexOf(arg.operator) !== -1;
        }

        function isNegative(arg) {
            return arg.type === "BinaryExpression" &&
                ["!==", "!="].indexOf(arg.operator) !== -1;
        }

        function getSuggestedAssertion(criteria) {
            const assertVar = getAssertContextVar();
            let assertMethod;

            if (criteria.strict) {
                assertMethod = criteria.negative ? "notStrictEqual" : "strictEqual";
            } else {
                assertMethod = criteria.negative ? "notEqual" : "equal";
            }

            if (assertVar) {
                return `${assertVar}.${assertMethod}`;
            }

            return assertMethod;
        }

        function checkArguments(args, isCalleeNegative, isGlobal, node) {
            /* istanbul ignore else: will correctly do nothing */
            if (args.length) {
                const firstArg = args[0],
                    isArgEqual = isEqual(firstArg),
                    isArgStrictEqual = isStrict(firstArg),
                    isArgNegative = isNegative(firstArg);

                if (isArgEqual) {
                    context.report({
                        node: node,
                        messageId: "noEqualityCheckInOk",
                        data: {
                            assertion: sourceCode.getText(node.callee),
                            suggestion: getSuggestedAssertion({
                                strict: isArgStrictEqual,
                                negative: isArgNegative !== isCalleeNegative
                            }),
                            a: sourceCode.getText(firstArg.left),
                            b: sourceCode.getText(firstArg.right)
                        }
                    });
                }
            }
        }

        return {
            "CallExpression": function (node) {
                if (asyncStateStack.length && isOkOrNotOk(node.callee)) {
                    const isGlobal = node.callee.type === "Identifier";
                    checkArguments(node.arguments, isNotOk(node.callee), isGlobal, node);
                } else if (utils.isTest(node.callee)) {
                    asyncStateStack.push({
                        assertContextVar: utils.getAssertContextNameForTest(node.arguments)
                    });
                }
            },

            "CallExpression:exit": function (node) {
                if (utils.isTest(node.callee)) {
                    asyncStateStack.pop();
                }
            }
        };
    }
};
